@@ -55,9 +55,17 @@ class DesynchronizedError(VirlException):
     pass
 
 
+class InvalidContentType(VirlException):
+    pass
+
+
 class InvalidImageFile(VirlException):
     pass
 
 
 class InvalidProperty(VirlException):
     pass
+
+
+class MethodNotActive(VirlException):
+    pass
