@@ -24,6 +24,8 @@ import io
 from typing import TYPE_CHECKING, Any, Optional
 
 if TYPE_CHECKING:
+    from pyats.topology import Device
+
     from .lab import Lab
 
 
@@ -42,8 +44,8 @@ class ClPyats:
         against a device either in exec mode ``show version`` or in
         configuration mode ``interface gi0/0 \\n no shut``.
 
-        :param lab: The lab which should be used with pyATS
-        :param hostname: Force hostname/ip and port for console terminal server
+        :param lab: the lab which should be used with pyATS
+        :param hostname: force hostname/ip and port for console terminal server
         :raises PyatsNotInstalled: when pyATS can not be found
         :raises PyatsDeviceNotFound: when the device can not be found
         """
@@ -59,6 +61,15 @@ class ClPyats:
         self._testbed: Any = None
         self._connections: list[Any] = []
 
+    @property
+    def hostname(self) -> Optional[str]:
+        """Return forced hostname/ip and port terminal server setting"""
+        return self._hostname
+
+    @hostname.setter
+    def hostname(self, hostname: Optional[str] = None) -> None:
+        self._hostname = hostname
+
     def _check_pyats_installed(self) -> None:
         if not self._pyats_installed:
             raise PyatsNotInstalled
@@ -82,16 +93,8 @@ class ClPyats:
         data.devices.terminal_server.credentials.default.password = password
         self._testbed = data
 
-    def run_command(self, node_label: str, command: str) -> str:
-        """
-        Run a command on the device in `exec` mode.
-
-        :param node_label: the label / title of the device
-        :param command: the command to be run in exec mode
-        :returns: The output from the device
-        """
+    def _prepare_device(self, node_label: str) -> Device:
         self._check_pyats_installed()
-
         try:
             pyats_device = self._testbed.devices[node_label]
         except KeyError:
@@ -101,29 +104,69 @@ class ClPyats:
         # TODO: later look at pooling connections
         pyats_device.connect(log_stdout=False, learn_hostname=True)
         self._connections.append(pyats_device)
-        return pyats_device.execute(command, log_stdout=False)
+        return pyats_device
 
-    def run_config_command(self, node_label: str, command: str) -> str:
+    def _prepare_params(
+        self,
+        init_exec_commands: Optional[list] = None,
+        init_config_commands: Optional[list] = None,
+    ) -> dict:
+        """
+        Prepare a dictionary of optional parameters to be executed before a command.
+        None means that default commands will be executed.  If you want no commands
+        to be executed pass an empty list instead.
+
+        :param init_exec_commands: a list of exec commands to be executed
+        :param init_config_commangs: a list of config commands to be executed
+        :returns:  a dictionary of optional parameters to be executed with a command
+        """
+        params = {}
+        if init_exec_commands:
+            params["init_exec_commands"] = init_exec_commands
+        if init_config_commands:
+            params["init_config_commands"] = init_config_commands
+        return params
+
+    def run_command(
+        self,
+        node_label: str,
+        command: str,
+        init_exec_commands: Optional[list] = None,
+        init_config_commands: Optional[list] = None,
+    ) -> str:
+        """
+        Run a command on the device in `exec` mode.
+
+        :param node_label: the label / title of the device
+        :param command: the command to be run in exec mode
+        :param init_exec_commands: a list of exec commands to be executed
+        :param init_config_commangs: a list of config commands to be executed
+        :returns: the output from the device
+        """
+        pyats_device = self._prepare_device(node_label)
+        params = self._prepare_params(init_exec_commands, init_config_commands)
+        return pyats_device.execute(command, log_stdout=False, **params)
+
+    def run_config_command(
+        self,
+        node_label: str,
+        command: str,
+        init_exec_commands: Optional[list] = None,
+        init_config_commands: Optional[list] = None,
+    ) -> str:
         """
         Run a command on the device in `configure` mode. pyATS
         handles the change into `configure` mode automatically.
 
         :param node_label: the label / title of the device
         :param command: the command to be run in exec mode
-        :returns: The output from the device
+        :param init_exec_commands: a list of exec commands to be executed
+        :param init_config_commangs: a list of config commands to be executed
+        :returns: the output from the device
         """
-        self._check_pyats_installed()
-
-        try:
-            pyats_device = self._testbed.devices[node_label]
-        except KeyError:
-            raise PyatsDeviceNotFound(node_label)
-
-        # TODO: later check if connected
-        # TODO: later look at pooling connections
-        pyats_device.connect(log_stdout=False, learn_hostname=True)
-        self._connections.append(pyats_device)
-        return pyats_device.configure(command, log_stdout=False)
+        pyats_device = self._prepare_device(node_label)
+        params = self._prepare_params(init_exec_commands, init_config_commands)
+        return pyats_device.configure(command, log_stdout=False, **params)
 
     def cleanup(self) -> None:
         """Cleans up the pyATS connections."""
